# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Deprecation utilities."""

__docformat__ = "restructuredtext en"

import os
import sys
from enum import Enum
from warnings import warn
from functools import WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES
from typing import Any, Callable, Dict, Optional, Type
from typing_extensions import Protocol


def get_real__name__(some_callable: Callable) -> str:
    """
    This is another super edge magic case which is needed because we uses
    lazy_wraps because of logilab.common.modutils.LazyObject and because
    __name__ has special behavior and doesn't work like a normal attribute and
    that __getattribute__ of lazy_wraps is bypassed.

    Therefor, to get the real callable name when several lazy_wrapped
    decorator are used we need to travers the __wrapped__ attributes chain.
    """

    targeted_callable = some_callable
    while hasattr(targeted_callable, "__wrapped__"):
        targeted_callable = targeted_callable.__wrapped__  # type: ignore

    return targeted_callable.__name__


def lazy_wraps(wrapped: Callable) -> Callable:
    """
    This is the equivalent of the @wraps decorator of functools except it won't
    try to grabs attributes of the targeted function on decoration but on access.

    This is needed because of logilab.common.modutils.LazyObject.

    Indeed: if you try to decorate a LazyObject with @wraps, wraps will try to
    access attributes of LazyObject and this will trigger the attempt to import
    the module decorated by LazyObject which you don't want to do when you just
    want to mark this LazyObject has been a deprecated objet that you only
    wants to trigger if the user try to use it.

    Usage: like @wraps()

    >>> @lazy_wraps(function)
    >>> def wrapper(*args, **kwargs): ...
    """

    def update_wrapper_attributes(wrapper: Callable) -> Callable:
        def __getattribute__(self, attribute: str) -> Any:
            if attribute in WRAPPER_ASSIGNMENTS:
                return getattr(wrapped, attribute)

            return super(self.__class__, self).__getattribute__(attribute)

        wrapper.__getattribute__ = __getattribute__  # type: ignore

        for attribute in WRAPPER_UPDATES:
            getattr(wrapper, attribute).update(getattr(wrapped, attribute, {}))

        wrapper.__wrapped__ = wrapped  # type: ignore

        return wrapper

    return update_wrapper_attributes


class DeprecationWrapper(object):
    """proxy to print a warning on access to any attribute of the wrapped object"""

    def __init__(
        self, proxied: Any, msg: Optional[str] = None, version: Optional[str] = None
    ) -> None:
        self._proxied: Any = proxied
        self._msg: str = msg if msg else ""
        self.version: Optional[str] = version

    def __getattr__(self, attr: str) -> Any:
        send_warning(
            self._msg,
            deprecation_class=DeprecationWarning,
            deprecation_class_kwargs={},
            stacklevel=3,
            version=self.version,
        )
        return getattr(self._proxied, attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in ("_proxied", "_msg"):
            self.__dict__[attr] = value
        else:
            send_warning(
                self._msg,
                deprecation_class=DeprecationWarning,
                deprecation_class_kwargs={},
                stacklevel=3,
                version=self.version,
            )
            setattr(self._proxied, attr, value)


def _get_module_name(number: int = 1) -> str:
    """
    automagically try to determine the package name from which the warning has
    been triggered by loop other calling frames.

    If it fails to do so, return an empty string.
    """

    frame = sys._getframe()

    for i in range(number + 1):
        if frame.f_back is None:
            break

        frame = frame.f_back

    if frame.f_globals["__package__"]:
        return frame.f_globals["__package__"]

    file_name = os.path.split(frame.f_globals["__file__"])[1]

    if file_name.endswith(".py"):
        file_name = file_name[: -len(".py")]

    return file_name


def send_warning(
    reason: str,
    deprecation_class: Type[DeprecationWarning],
    deprecation_class_kwargs: Dict[str, Any],
    version: Optional[str] = None,
    stacklevel: int = 2,
    module_name: Optional[str] = None,
) -> None:
    """Display a deprecation message only if the version is older than the
    compatible version.
    """
    if module_name and version:
        reason = "[%s %s] %s" % (module_name, version, reason)
    elif module_name:
        reason = "[%s] %s" % (module_name, reason)
    elif version:
        reason = "[%s] %s" % (version, reason)

    warn(
        deprecation_class(reason, **deprecation_class_kwargs), stacklevel=stacklevel  # type: ignore
    )


class DeprecationWarningKind(Enum):
    ARGUMENT = "argument"
    ATTRIBUTE = "attribute"
    CALLABLE = "callable"
    CLASS = "class"
    MODULE = "module"


class DeprecationWarningOperation(Enum):
    DEPRECATED = "deprecated"
    MOVED = "moved"
    REMOVED = "removed"
    RENAMED = "renamed"


class StructuredDeprecationWarning(DeprecationWarning):
    """
    Base class for all structured DeprecationWarning
    Mostly used with isinstance
    """

    def __init__(self, reason: str):
        self.reason: str = reason

    def __str__(self) -> str:
        return self.reason


class TargetRenamedDeprecationWarning(StructuredDeprecationWarning):
    def __init__(self, reason: str, kind: DeprecationWarningKind, old_name: str, new_name: str):
        super().__init__(reason)
        self.operation = DeprecationWarningOperation.RENAMED
        self.kind: DeprecationWarningKind = kind  # callable, class, module, argument, attribute
        self.old_name: str = old_name
        self.new_name: str = new_name


class TargetDeprecatedDeprecationWarning(StructuredDeprecationWarning):
    def __init__(self, reason: str, kind: DeprecationWarningKind):
        super().__init__(reason)
        self.operation = DeprecationWarningOperation.DEPRECATED
        self.kind: DeprecationWarningKind = kind  # callable, class, module, argument, attribute


class TargetRemovedDeprecationWarning(StructuredDeprecationWarning):
    def __init__(self, reason: str, kind: DeprecationWarningKind, name: str):
        super().__init__(reason)
        self.operation = DeprecationWarningOperation.REMOVED
        self.kind: DeprecationWarningKind = kind  # callable, class, module, argument, attribute
        self.name: str = name


class TargetMovedDeprecationWarning(StructuredDeprecationWarning):
    def __init__(
        self,
        reason: str,
        kind: DeprecationWarningKind,
        old_name: str,
        new_name: str,
        old_module: str,
        new_module: str,
    ):
        super().__init__(reason)
        self.operation = DeprecationWarningOperation.MOVED
        self.kind: DeprecationWarningKind = kind  # callable, class, module, argument, attribute
        self.old_name: str = old_name
        self.new_name: str = new_name
        self.old_module: str = old_module
        self.new_module: str = new_module


def callable_renamed(
    old_name: str, new_function: Callable, version: Optional[str] = None
) -> Callable:
    """use to tell that a callable has been renamed.

    It returns a callable wrapper, so that when its called a warning is printed
    telling what is the object new name.

    >>> old_function = renamed('old_function', new_function)
    >>> old_function()
    sample.py:57: DeprecationWarning: old_function has been renamed and is deprecated, uses
    new_function instead old_function()
    >>>
    """

    @lazy_wraps(new_function)
    def wrapped(*args, **kwargs):
        send_warning(
            (
                f"{old_name} has been renamed and is deprecated, uses "
                f"{get_real__name__(new_function)} instead"
            ),
            TargetRenamedDeprecationWarning,
            deprecation_class_kwargs={
                "kind": DeprecationWarningKind.CALLABLE,
                "old_name": old_name,
                "new_name": get_real__name__(new_function),
            },
            stacklevel=3,
            version=version,
            module_name=new_function.__module__,
        )
        return new_function(*args, **kwargs)

    return wrapped


renamed: Callable[[str, Callable, Optional[str]], Callable] = callable_renamed(
    old_name="renamed", new_function=callable_renamed
)


def argument_removed(old_argument_name: str, version: Optional[str] = None) -> Callable:
    """
    callable decorator to allow getting backward compatibility for renamed keyword arguments.

    >>> @argument_removed("old")
    ... def some_function(new):
    ...     return new
    >>> some_function(old=42)
    sample.py:15: DeprecationWarning: argument old of callable some_function has been renamed and
    is deprecated, use keyword argument new instead some_function(old=42)
    42
    """

    def _wrap(func: Callable) -> Callable:
        @lazy_wraps(func)
        def check_kwargs(*args, **kwargs):
            if old_argument_name in kwargs:
                send_warning(
                    f"argument {old_argument_name} of callable {get_real__name__(func)} has been "
                    f"removed and is deprecated",
                    deprecation_class=TargetRemovedDeprecationWarning,
                    deprecation_class_kwargs={
                        "kind": DeprecationWarningKind.ARGUMENT,
                        "name": old_argument_name,
                    },
                    stacklevel=3,
                    version=version,
                    module_name=func.__module__,
                )
                del kwargs[old_argument_name]

            return func(*args, **kwargs)

        return check_kwargs

    return _wrap


@argument_removed("name")
@argument_removed("doc")
def callable_deprecated(
    reason: Optional[str] = None, version: Optional[str] = None, stacklevel: int = 2
) -> Callable:
    """Display a deprecation message only if the version is older than the
    compatible version.
    """

    def decorator(func: Callable) -> Callable:
        @lazy_wraps(func)
        def wrapped(*args, **kwargs) -> Callable:
            message: str = reason or 'The function "%s" is deprecated'
            if "%s" in message:
                message %= get_real__name__(func)

            send_warning(
                message,
                TargetDeprecatedDeprecationWarning,
                {"kind": DeprecationWarningKind.CALLABLE},
                version,
                stacklevel + 1,
                module_name=func.__module__,
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator


class CallableDeprecatedCallable(Protocol):
    def __call__(
        self, reason: Optional[str] = None, version: Optional[str] = None, stacklevel: int = 2
    ) -> Callable:
        ...


deprecated: CallableDeprecatedCallable = callable_renamed(
    old_name="deprecated", new_function=callable_deprecated
)


def _generate_class_deprecated():
    class _class_deprecated(type):
        """metaclass to print a warning on instantiation of a deprecated class"""

        def __call__(cls, *args, **kwargs):
            message = getattr(cls, "__deprecation_warning__", "%(cls)s is deprecated") % {
                "cls": cls.__name__
            }
            send_warning(
                message,
                deprecation_class=getattr(
                    cls, "__deprecation_warning_class__", TargetDeprecatedDeprecationWarning
                ),
                deprecation_class_kwargs=getattr(
                    cls,
                    "__deprecation_warning_class_kwargs__",
                    {"kind": DeprecationWarningKind.CLASS},
                ),
                module_name=getattr(
                    cls, "__deprecation_warning_module_name__", _get_module_name(1)
                ),
                stacklevel=getattr(cls, "__deprecation_warning_stacklevel__", 3),
                version=getattr(cls, "__deprecation_warning_version__", None),
            )
            return type.__call__(cls, *args, **kwargs)

    return _class_deprecated


class_deprecated = _generate_class_deprecated()


def attribute_renamed(old_name: str, new_name: str, version: Optional[str] = None) -> Callable:
    """
    class decorator to allow getting backward compatibility for renamed attributes.

    >>> @attribute_renamed(old_name="old", new_name="new")
    ... class SomeClass:
    ...     def __init__(self):
    ...         self.new = 42

    >>> some_class = SomeClass()
    >>> print(some_class.old)
    sample.py:15: DeprecationWarning: SomeClass.old has been renamed and is deprecated, use
    SomeClass.new instead
      print(some_class.old)
    42
    >>> some_class.old = 43
    sample.py:16: DeprecationWarning: SomeClass.old has been renamed and is deprecated, use
    SomeClass.new instead
      some_class.old = 43
    >>> some_class.old == some_class.new
    True
    """

    def _class_wrap(klass: type) -> type:
        reason = (
            f"{klass.__name__}.{old_name} has been renamed and is deprecated, use "
            f"{klass.__name__}.{new_name} instead"
        )

        def _get_old(self) -> Any:
            send_warning(
                reason,
                deprecation_class=TargetRenamedDeprecationWarning,
                deprecation_class_kwargs={
                    "kind": DeprecationWarningKind.ATTRIBUTE,
                    "old_name": old_name,
                    "new_name": new_name,
                },
                stacklevel=3,
                version=version,
                module_name=klass.__module__,
            )
            return getattr(self, new_name)

        def _set_old(self, value) -> None:
            send_warning(
                reason,
                deprecation_class=TargetRenamedDeprecationWarning,
                deprecation_class_kwargs={
                    "kind": DeprecationWarningKind.ATTRIBUTE,
                    "old_name": old_name,
                    "new_name": new_name,
                },
                stacklevel=3,
                version=version,
                module_name=klass.__module__,
            )
            setattr(self, new_name, value)

        def _del_old(self):
            send_warning(
                reason,
                deprecation_class=TargetRenamedDeprecationWarning,
                deprecation_class_kwargs={
                    "kind": DeprecationWarningKind.ATTRIBUTE,
                    "old_name": old_name,
                    "new_name": new_name,
                },
                stacklevel=3,
                version=version,
                module_name=klass.__module__,
            )
            delattr(self, new_name)

        setattr(klass, old_name, property(_get_old, _set_old, _del_old))

        return klass

    return _class_wrap


def argument_renamed(old_name: str, new_name: str, version: Optional[str] = None) -> Callable:
    """
    callable decorator to allow getting backward compatibility for renamed keyword arguments.

    >>> @argument_renamed(old_name="old", new_name="new")
    ... def some_function(new):
    ...     return new
    >>> some_function(old=42)
    sample.py:15: DeprecationWarning: argument old of callable some_function has been renamed and
    is deprecated, use keyword argument new instead
      some_function(old=42)
    42
    """

    def _wrap(func: Callable) -> Callable:
        @lazy_wraps(func)
        def check_kwargs(*args, **kwargs) -> Callable:
            if old_name in kwargs and new_name in kwargs:
                raise ValueError(
                    f"argument {old_name} of callable {get_real__name__(func)} has been "
                    f"renamed to {new_name} but you are both using {old_name} and "
                    f"{new_name} has keyword arguments, only uses {new_name}"
                )

            if old_name in kwargs:
                send_warning(
                    f"argument {old_name} of callable {get_real__name__(func)} has been renamed "
                    f"and is deprecated, use keyword argument {new_name} instead",
                    deprecation_class=TargetRenamedDeprecationWarning,
                    deprecation_class_kwargs={
                        "kind": DeprecationWarningKind.ARGUMENT,
                        "old_name": old_name,
                        "new_name": new_name,
                    },
                    stacklevel=3,
                    version=version,
                    module_name=func.__module__,
                )
                kwargs[new_name] = kwargs[old_name]
                del kwargs[old_name]

            return func(*args, **kwargs)

        return check_kwargs

    return _wrap


@argument_renamed(old_name="modpath", new_name="module_path")
@argument_renamed(old_name="objname", new_name="object_name")
def callable_moved(
    module_name: str,
    object_name: str,
    version: Optional[str] = None,
    stacklevel: int = 2,
    new_name: Optional[str] = None,
) -> Callable:
    """use to tell that a callable has been moved to a new module.

    It returns a callable wrapper, so that when its called a warning is printed
    telling where the object can be found, import is done (and not before) and
    the actual object is called.

    NOTE: the usage is somewhat limited on classes since it will fail if the
    wrapper is use in a class ancestors list, use the `class_moved` function
    instead (which has no lazy import feature though).
    """
    # in case the callable has been renamed
    new_name = new_name if new_name is not None else object_name
    old_module = _get_module_name(3)

    message = "object %s.%s has been moved to %s.%s" % (
        old_module,
        object_name,
        module_name,
        object_name,
    )

    def callnew(*args, **kwargs):
        from logilab.common.modutils import load_module_from_name

        send_warning(
            message,
            deprecation_class=TargetMovedDeprecationWarning,
            deprecation_class_kwargs={
                "kind": DeprecationWarningKind.CALLABLE,
                "old_name": object_name,
                "new_name": new_name,
                "old_module": old_module,
                "new_module": module_name,
            },
            version=version,
            stacklevel=stacklevel + 1,
            module_name=old_module,
        )

        m = load_module_from_name(module_name)
        return getattr(m, object_name)(*args, **kwargs)

    return callnew


moved: Callable[[Optional[str], Optional[str], int], Callable] = callable_renamed(
    old_name="moved", new_function=callable_moved
)


def class_renamed(
    old_name: str,
    new_class: type,
    message: Optional[str] = None,
    version: Optional[str] = None,
    module_name: Optional[str] = None,
    deprecated_warning_class=TargetRenamedDeprecationWarning,
    deprecated_warning_kwargs=None,
) -> type:
    """automatically creates a class which fires a DeprecationWarning
    when instantiated.

    >>> Set = class_renamed('Set', set, 'Set is now replaced by set')
    >>> s = Set()
    sample.py:57: DeprecationWarning: Set is now replaced by set
    s = Set()
    >>>
    """
    class_dict: Dict[str, Any] = {}
    if message is None:
        message = "%s is deprecated, use %s instead" % (old_name, new_class.__name__)

    class_dict["__deprecation_warning__"] = message
    class_dict["__deprecation_warning_class__"] = deprecated_warning_class
    if deprecated_warning_kwargs is None:
        class_dict["__deprecation_warning_class_kwargs__"] = {
            "kind": DeprecationWarningKind.CLASS,
            "old_name": old_name,
            "new_name": new_class.__name__,
        }
    else:
        class_dict["__deprecation_warning_class_kwargs__"] = deprecated_warning_kwargs
    class_dict["__deprecation_warning_version__"] = version
    class_dict["__deprecation_warning_stacklevel__"] = 3

    if module_name:
        class_dict["__deprecation_warning_module_name__"] = module_name
    else:
        class_dict["__deprecation_warning_module_name__"] = _get_module_name(1)

    try:
        return class_deprecated(old_name, (new_class,), class_dict)
    except (NameError, TypeError):
        # in case of conflicting metaclass situation
        # mypy can't handle dynamic base classes https://github.com/python/mypy/issues/2477
        class DeprecatedClass(new_class):  # type: ignore
            def __init__(self, *args, **kwargs):
                msg = class_dict.get(
                    "__deprecation_warning__",
                    f"{old_name} is deprecated, use {new_class.__name__} instead",
                )
                send_warning(
                    msg,
                    deprecation_class=TargetRenamedDeprecationWarning,
                    deprecation_class_kwargs={
                        "kind": DeprecationWarningKind.CLASS,
                        "old_name": old_name,
                        "new_name": new_class.__name__,
                    },
                    stacklevel=class_dict.get("__deprecation_warning_stacklevel__", 3),
                    version=class_dict.get("__deprecation_warning_version__", None),
                )
                super(DeprecatedClass, self).__init__(*args, **kwargs)

        return DeprecatedClass


def class_moved(
    new_class: type,
    old_name: Optional[str] = None,
    message: Optional[str] = None,
    version: Optional[str] = None,
) -> type:
    """nice wrapper around class_renamed when a class has been moved into
    another module
    """
    if old_name is None:
        old_name = new_class.__name__

    old_module = _get_module_name(1)

    if message is None:
        message = "class %s.%s is now available as %s.%s" % (
            old_module,
            old_name,
            new_class.__module__,
            new_class.__name__,
        )

    module_name = _get_module_name(1)

    return class_renamed(
        old_name,
        new_class,
        message=message,
        module_name=module_name,
        deprecated_warning_class=TargetMovedDeprecationWarning,
        deprecated_warning_kwargs={
            "kind": DeprecationWarningKind.CLASS,
            "old_module": old_module,
            "new_module": new_class.__module__,
            "old_name": old_name,
            "new_name": new_class.__name__,
        },
    )



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Add an abstraction level to transparently import optik classes from optparse
(python >= 2.3) or the optik package.

It also defines three new types for optik/optparse command line parser :

  * regexp
    argument of this type will be converted using re.compile
  * csv
    argument of this type will be converted using split(',')
  * yn
    argument of this type will be true if 'y' or 'yes', false if 'n' or 'no'
  * named
    argument of this type are in the form <NAME>=<VALUE> or <NAME>:<VALUE>
  * password
    argument of this type wont be converted but this is used by other tools
    such as interactive prompt for configuration to double check value and
    use an invisible field
  * multiple_choice
    same as default "choice" type but multiple choices allowed
  * file
    argument of this type wont be converted but checked that the given file exists
  * color
    argument of this type wont be converted but checked its either a
    named color or a color specified using hexadecimal notation (preceded by a #)
  * time
    argument of this type will be converted to a float value in seconds
    according to time units (ms, s, min, h, d)
  * bytes
    argument of this type will be converted to a float value in bytes
    according to byte units (b, kb, mb, gb, tb)
"""
from __future__ import print_function

__docformat__ = "restructuredtext en"

import re
import sys
import time
from copy import copy
from os.path import exists
from logilab.common import attrdict

from typing import Any, Union, List, Optional, Tuple, Dict
from _io import StringIO

# python >= 2.3
from optparse import (  # noqa
    OptionParser as BaseParser,
    Option as BaseOption,
    OptionGroup,
    OptionContainer,
    OptionValueError,
    OptionError,
    Values,
    HelpFormatter,
    SUPPRESS_HELP,
    NO_DEFAULT,
)

try:
    from mx import DateTime

    HAS_MX_DATETIME = True
except ImportError:
    HAS_MX_DATETIME = False

from logilab.common.textutils import splitstrip, TIME_UNITS, BYTE_UNITS, apply_units


def check_regexp(option, opt, value):
    """check a regexp value by trying to compile it
    return the compiled regexp
    """
    if hasattr(value, "pattern"):
        return value
    try:
        return re.compile(value)
    except ValueError:
        raise OptionValueError("option %s: invalid regexp value: %r" % (opt, value))


def check_csv(
    option: Optional["Option"], opt: str, value: Union[List[str], Tuple[str, ...], str]
) -> Union[List[str], Tuple[str, ...]]:
    """check a csv value by trying to split it
    return the list of separated values
    """
    if isinstance(value, (list, tuple)):
        return value
    try:
        return splitstrip(value)
    except ValueError:
        raise OptionValueError("option %s: invalid csv value: %r" % (opt, value))


def check_yn(option: Optional["Option"], opt: str, value: Union[bool, str]) -> bool:
    """check a yn value
    return true for yes and false for no
    """
    if isinstance(value, int):
        return bool(value)
    if value in ("y", "yes"):
        return True
    if value in ("n", "no"):
        return False
    msg = "option %s: invalid yn value %r, should be in (y, yes, n, no)"
    raise OptionValueError(msg % (opt, value))


def check_named(
    option: Optional[Any], opt: str, value: Union[Dict[str, str], str]
) -> Dict[str, str]:
    """check a named value
    return a dictionary containing (name, value) associations
    """
    if isinstance(value, dict):
        return value
    values: List[Tuple[str, str]] = []
    for value in check_csv(option, opt, value):
        # mypy: Argument 1 to "append" of "list" has incompatible type "List[str]";
        # mypy: expected "Tuple[str, str]"
        # we know that the split will give a 2 items list
        if value.find("=") != -1:
            values.append(value.split("=", 1))  # type: ignore
        elif value.find(":") != -1:
            values.append(value.split(":", 1))  # type: ignore
    if values:
        return dict(values)
    msg = "option %s: invalid named value %r, should be <NAME>=<VALUE> or \
<NAME>:<VALUE>"
    raise OptionValueError(msg % (opt, value))


def check_password(option, opt, value):
    """check a password value (can't be empty)"""
    # no actual checking, monkey patch if you want more
    return value


def check_file(option, opt, value):
    """check a file value
    return the filepath
    """
    if exists(value):
        return value
    msg = "option %s: file %r does not exist"
    raise OptionValueError(msg % (opt, value))


# XXX use python datetime
def check_date(option, opt, value):
    """check a file value
    return the filepath
    """
    try:
        return DateTime.strptime(value, "%Y/%m/%d")
    except DateTime.Error:
        raise OptionValueError("expected format of %s is yyyy/mm/dd" % opt)


def check_color(option, opt, value):
    """check a color value and returns it
    /!\\ does *not* check color labels (like 'red', 'green'), only
    checks hexadecimal forms
    """
    # Case (1) : color label, we trust the end-user
    if re.match("[a-z0-9 ]+$", value, re.I):
        return value
    # Case (2) : only accepts hexadecimal forms
    if re.match("#[a-f0-9]{6}", value, re.I):
        return value
    # Else : not a color label neither a valid hexadecimal form => error
    msg = "option %s: invalid color : %r, should be either hexadecimal \
    value or predefined color"
    raise OptionValueError(msg % (opt, value))


def check_time(option, opt, value):
    if isinstance(value, (int, float)):
        return value
    return apply_units(value, TIME_UNITS)


def check_bytes(option: Optional["Option"], opt: str, value: Any) -> int:
    if hasattr(value, "__int__"):
        return value
    # mypy: Incompatible return value type (got "Union[float, int]", expected "int")
    # we force "int" using "final=int"
    return apply_units(value, BYTE_UNITS, final=int)  # type: ignore


class Option(BaseOption):
    """override optik.Option to add some new option types"""

    TYPES = BaseOption.TYPES + (
        "regexp",
        "csv",
        "yn",
        "named",
        "password",
        "multiple_choice",
        "file",
        "color",
        "time",
        "bytes",
    )
    ATTRS = BaseOption.ATTRS + ["hide", "level"]
    TYPE_CHECKER = copy(BaseOption.TYPE_CHECKER)
    TYPE_CHECKER["regexp"] = check_regexp
    TYPE_CHECKER["csv"] = check_csv
    TYPE_CHECKER["yn"] = check_yn
    TYPE_CHECKER["named"] = check_named
    TYPE_CHECKER["multiple_choice"] = check_csv
    TYPE_CHECKER["file"] = check_file
    TYPE_CHECKER["color"] = check_color
    TYPE_CHECKER["password"] = check_password
    TYPE_CHECKER["time"] = check_time
    TYPE_CHECKER["bytes"] = check_bytes
    if HAS_MX_DATETIME:
        TYPES += ("date",)
        TYPE_CHECKER["date"] = check_date

    def __init__(self, *opts: str, **attrs: Any) -> None:
        BaseOption.__init__(self, *opts, **attrs)
        # mypy: "Option" has no attribute "hide"
        # we test that in the if
        if hasattr(self, "hide") and self.hide:  # type: ignore
            self.help = SUPPRESS_HELP

    def _check_choice(self) -> None:
        """FIXME: need to override this due to optik misdesign"""
        if self.type in ("choice", "multiple_choice"):
            # mypy: "Option" has no attribute "choices"
            # we know that option of this type has this attribute
            if self.choices is None:  # type: ignore
                raise OptionError("must supply a list of choices for type 'choice'", self)
            elif not isinstance(self.choices, (tuple, list)):  # type: ignore
                raise OptionError(
                    "choices must be a list of strings ('%s' supplied)"
                    % str(type(self.choices)).split("'")[1],  # type: ignore
                    self,
                )
        elif self.choices is not None:  # type: ignore
            raise OptionError("must not supply choices for type %r" % self.type, self)

    # mypy: Unsupported target for indexed assignment
    # black magic?
    BaseOption.CHECK_METHODS[2] = _check_choice  # type: ignore

    def process(self, opt: str, value: str, values: Values, parser: BaseParser) -> int:
        # First, convert the value(s) to the right type.  Howl if any
        # value(s) are bogus.
        value = self.convert_value(opt, value)
        if self.type == "named":
            assert self.dest is not None
            existant = getattr(values, self.dest)
            if existant:
                existant.update(value)
                value = existant
        # And then take whatever action is expected of us.
        # This is a separate method to make life easier for
        # subclasses to add new actions.
        # mypy: Argument 2 to "take_action" of "Option" has incompatible type "Optional[str]";
        # mypy: expected "str"
        # is it ok?
        return self.take_action(self.action, self.dest, opt, value, values, parser)  # type: ignore


class OptionParser(BaseParser):
    """override optik.OptionParser to use our Option class"""

    def __init__(self, option_class: type = Option, *args: Any, **kwargs: Any) -> None:
        # mypy: Argument "option_class" to "__init__" of "OptionParser" has incompatible type
        # mypy: "type"; expected "Option"
        # mypy is doing really weird things with *args/**kwargs and looks buggy
        BaseParser.__init__(self, option_class=option_class, *args, **kwargs)  # type: ignore

    def format_option_help(self, formatter: Optional[HelpFormatter] = None) -> str:
        if formatter is None:
            formatter = self.formatter
        outputlevel = getattr(formatter, "output_level", 0)
        formatter.store_option_strings(self)
        result = []
        result.append(formatter.format_heading("Options"))
        formatter.indent()
        if self.option_list:
            result.append(OptionContainer.format_option_help(self, formatter))
            result.append("\n")
        for group in self.option_groups:
            # mypy: "OptionParser" has no attribute "level"
            # but it has one no?
            if group.level <= outputlevel and (  # type: ignore
                group.description or level_options(group, outputlevel)
            ):
                result.append(group.format_help(formatter))
                result.append("\n")
        formatter.dedent()
        # Drop the last "\n", or the header if no options or option groups:
        return "".join(result[:-1])


# mypy error: error: "Type[OptionGroup]" has no attribute "level"
# monkeypatching
OptionGroup.level = 0  # type: ignore


def level_options(group: BaseParser, outputlevel: int) -> List[BaseOption]:
    # mypy: "Option" has no attribute "help"
    # but it does
    return [
        option
        for option in group.option_list
        if (getattr(option, "level", 0) or 0) <= outputlevel and option.help is not SUPPRESS_HELP  # type: ignore  # noqa
    ]


def format_option_help(self, formatter):
    result = []
    outputlevel = getattr(formatter, "output_level", 0) or 0
    for option in level_options(self, outputlevel):
        result.append(formatter.format_option(option))
    return "".join(result)


# mypy error: Cannot assign to a method
# but we still do it because magic
OptionContainer.format_option_help = format_option_help  # type: ignore


class ManHelpFormatter(HelpFormatter):
    """Format help using man pages ROFF format"""

    def __init__(
        self,
        indent_increment: int = 0,
        max_help_position: int = 24,
        width: int = 79,
        short_first: int = 0,
    ) -> None:
        HelpFormatter.__init__(self, indent_increment, max_help_position, width, short_first)

    def format_heading(self, heading: str) -> str:
        return ".SH %s\n" % heading.upper()

    def format_description(self, description):
        return description

    def format_option(self, option: BaseParser) -> str:
        try:
            # mypy: "Option" has no attribute "option_strings"
            # we handle if it doesn't
            optstring = option.option_strings  # type: ignore
        except AttributeError:
            optstring = self.format_option_strings(option)
        # mypy: "OptionParser" has no attribute "help"
        # it does
        if option.help:  # type: ignore
            # mypy: Argument 1 to "expand_default" of "HelpFormatter" has incompatible type
            # mypy: "OptionParser"; expected "Option"
            # it still works?
            help_text = self.expand_default(option)  # type: ignore
            help = " ".join([line.strip() for line in help_text.splitlines()])
        else:
            help = ""
        return """.IP "%s"
%s
""" % (
            optstring,
            help,
        )

    def format_head(self, optparser: OptionParser, pkginfo: attrdict, section: int = 1) -> str:
        long_desc = ""
        pgm = optparser.get_prog_name()
        short_desc = self.format_short_description(pgm, pkginfo.description)
        if hasattr(pkginfo, "long_desc"):
            long_desc = self.format_long_description(pgm, pkginfo.long_desc)
        return "%s\n%s\n%s\n%s" % (
            self.format_title(pgm, section),
            short_desc,
            self.format_synopsis(pgm),
            long_desc,
        )

    def format_title(self, pgm: str, section: int) -> str:
        date = "-".join([str(num) for num in time.localtime()[:3]])
        return '.TH %s %s "%s" %s' % (pgm, section, date, pgm)

    def format_short_description(self, pgm: str, short_desc: str) -> str:
        return r""".SH NAME
.B %s
\- %s
""" % (
            pgm,
            short_desc.strip(),
        )

    def format_synopsis(self, pgm: str) -> str:
        return (
            """.SH SYNOPSIS
.B  %s
[
.I OPTIONS
] [
.I <arguments>
]
"""
            % pgm
        )

    def format_long_description(self, pgm, long_desc):
        long_desc = "\n".join([line.lstrip() for line in long_desc.splitlines()])
        long_desc = long_desc.replace("\n.\n", "\n\n")
        if long_desc.lower().startswith(pgm):
            long_desc = long_desc[len(pgm) :]
        return """.SH DESCRIPTION
.B %s
%s
""" % (
            pgm,
            long_desc.strip(),
        )

    def format_tail(self, pkginfo: attrdict) -> str:
        tail = """.SH SEE ALSO
/usr/share/doc/pythonX.Y-%s/

.SH BUGS
Please report bugs on the project\'s mailing list:
%s

.SH AUTHOR
%s <%s>
""" % (
            getattr(pkginfo, "debian_name", pkginfo.modname),
            pkginfo.mailinglist,
            pkginfo.author,
            pkginfo.author_email,
        )

        if hasattr(pkginfo, "copyright"):
            tail += (
                """
.SH COPYRIGHT
%s
"""
                % pkginfo.copyright
            )

        return tail


def generate_manpage(
    optparser: OptionParser,
    pkginfo: attrdict,
    section: int = 1,
    stream: StringIO = sys.stdout,
    level: int = 0,
) -> None:
    """generate a man page from an optik parser"""
    formatter = ManHelpFormatter()
    # mypy: "ManHelpFormatter" has no attribute "output_level"
    # dynamic attribute?
    formatter.output_level = level  # type: ignore
    formatter.parser = optparser
    print(formatter.format_head(optparser, pkginfo, section), file=stream)
    print(optparser.format_option_help(formatter), file=stream)
    print(formatter.format_tail(pkginfo), file=stream)


__all__ = ("OptionParser", "Option", "OptionGroup", "OptionValueError", "Values")
